@@ -2,7 +2,8 @@ import time
 from decimal import Decimal
 
 import msgpack
-from eth_account.messages import encode_structured_data
+from eth_account import Account
+from eth_account.messages import encode_typed_data
 from eth_utils import keccak, to_hex
 
 from hyperliquid.utils.types import Cloid, Literal, NotRequired, Optional, TypedDict, Union
@@ -74,6 +75,55 @@ ScheduleCancelAction = TypedDict(
     },
 )
 
+USD_SEND_SIGN_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "destination", "type": "string"},
+    {"name": "amount", "type": "string"},
+    {"name": "time", "type": "uint64"},
+]
+
+SPOT_TRANSFER_SIGN_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "destination", "type": "string"},
+    {"name": "token", "type": "string"},
+    {"name": "amount", "type": "string"},
+    {"name": "time", "type": "uint64"},
+]
+
+WITHDRAW_SIGN_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "destination", "type": "string"},
+    {"name": "amount", "type": "string"},
+    {"name": "time", "type": "uint64"},
+]
+
+USD_CLASS_TRANSFER_SIGN_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "amount", "type": "string"},
+    {"name": "toPerp", "type": "bool"},
+    {"name": "nonce", "type": "uint64"},
+]
+
+TOKEN_DELEGATE_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "validator", "type": "address"},
+    {"name": "wei", "type": "uint64"},
+    {"name": "isUndelegate", "type": "bool"},
+    {"name": "nonce", "type": "uint64"},
+]
+
+CONVERT_TO_MULTI_SIG_USER_SIGN_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "signers", "type": "string"},
+    {"name": "nonce", "type": "uint64"},
+]
+
+MULTI_SIG_ENVELOPE_SIGN_TYPES = [
+    {"name": "hyperliquidChain", "type": "string"},
+    {"name": "multiSigActionHash", "type": "bytes32"},
+    {"name": "nonce", "type": "uint64"},
+]
+
 
 def order_type_to_wire(order_type: OrderType) -> OrderTypeWire:
     if "limit" in order_type:
@@ -93,7 +143,7 @@ def address_to_bytes(address):
     return bytes.fromhex(address[2:] if address.startswith("0x") else address)
 
 
-def action_hash(action, vault_address, nonce):
+def action_hash(action, vault_address, nonce, expires_after):
     data = msgpack.packb(action)
     data += nonce.to_bytes(8, "big")
     if vault_address is None:
@@ -101,6 +151,9 @@ def action_hash(action, vault_address, nonce):
     else:
         data += b"\x01"
         data += address_to_bytes(vault_address)
+    if expires_after is not None:
+        data += b"\x00"
+        data += expires_after.to_bytes(8, "big")
     return keccak(data)
 
 
@@ -108,10 +161,8 @@ def construct_phantom_agent(hash, is_mainnet):
     return {"source": "a" if is_mainnet else "b", "connectionId": hash}
 
 
-def sign_l1_action(wallet, action, active_pool, nonce, is_mainnet):
-    hash = action_hash(action, active_pool, nonce)
-    phantom_agent = construct_phantom_agent(hash, is_mainnet)
-    data = {
+def l1_payload(phantom_agent):
+    return {
         "domain": {
             "chainId": 1337,
             "name": "Exchange",
@@ -133,17 +184,15 @@ def sign_l1_action(wallet, action, active_pool, nonce, is_mainnet):
         "primaryType": "Agent",
         "message": phantom_agent,
     }
-    return sign_inner(wallet, data)
 
 
-def sign_user_signed_action(wallet, action, payload_types, primary_type, is_mainnet):
-    action["signatureChainId"] = "0x66eee"
-    action["hyperliquidChain"] = "Mainnet" if is_mainnet else "Testnet"
-    data = {
+def user_signed_payload(primary_type, payload_types, action):
+    chain_id = int(action["signatureChainId"], 16)
+    return {
         "domain": {
             "name": "HyperliquidSignTransaction",
             "version": "1",
-            "chainId": 421614,
+            "chainId": chain_id,
             "verifyingContract": "0x0000000000000000000000000000000000000000",
         },
         "types": {
@@ -158,19 +207,105 @@ def sign_user_signed_action(wallet, action, payload_types, primary_type, is_main
         "primaryType": primary_type,
         "message": action,
     }
+
+
+def sign_l1_action(wallet, action, active_pool, nonce, expires_after, is_mainnet):
+    hash = action_hash(action, active_pool, nonce, expires_after)
+    phantom_agent = construct_phantom_agent(hash, is_mainnet)
+    data = l1_payload(phantom_agent)
     return sign_inner(wallet, data)
 
 
+def sign_user_signed_action(wallet, action, payload_types, primary_type, is_mainnet):
+    # signatureChainId is the chain used by the wallet to sign and can be any chain.
+    # hyperliquidChain determines the environment and prevents replaying an action on a different chain.
+    action["signatureChainId"] = "0x66eee"
+    action["hyperliquidChain"] = "Mainnet" if is_mainnet else "Testnet"
+    data = user_signed_payload(primary_type, payload_types, action)
+    return sign_inner(wallet, data)
+
+
+def add_multi_sig_types(sign_types):
+    enriched_sign_types = []
+    enriched = False
+    for sign_type in sign_types:
+        enriched_sign_types.append(sign_type)
+        if sign_type["name"] == "hyperliquidChain":
+            enriched = True
+            enriched_sign_types.append(
+                {
+                    "name": "payloadMultiSigUser",
+                    "type": "address",
+                }
+            )
+            enriched_sign_types.append(
+                {
+                    "name": "outerSigner",
+                    "type": "address",
+                }
+            )
+    if not enriched:
+        print('"hyperliquidChain" missing from sign_types. sign_types was not enriched with multi-sig signing types')
+    return enriched_sign_types
+
+
+def add_multi_sig_fields(action, payload_multi_sig_user, outer_signer):
+    action = action.copy()
+    action["payloadMultiSigUser"] = payload_multi_sig_user.lower()
+    action["outerSigner"] = outer_signer.lower()
+    return action
+
+
+def sign_multi_sig_user_signed_action_payload(
+    wallet, action, is_mainnet, sign_types, tx_type, payload_multi_sig_user, outer_signer
+):
+    envelope = add_multi_sig_fields(action, payload_multi_sig_user, outer_signer)
+    sign_types = add_multi_sig_types(sign_types)
+    return sign_user_signed_action(
+        wallet,
+        envelope,
+        sign_types,
+        tx_type,
+        is_mainnet,
+    )
+
+
+def sign_multi_sig_l1_action_payload(
+    wallet, action, is_mainnet, vault_address, timestamp, expires_after, payload_multi_sig_user, outer_signer
+):
+    envelope = [payload_multi_sig_user.lower(), outer_signer.lower(), action]
+    return sign_l1_action(
+        wallet,
+        envelope,
+        vault_address,
+        timestamp,
+        expires_after,
+        is_mainnet,
+    )
+
+
+def sign_multi_sig_action(wallet, action, is_mainnet, vault_address, nonce, expires_after):
+    action_without_tag = action.copy()
+    del action_without_tag["type"]
+    multi_sig_action_hash = action_hash(action_without_tag, vault_address, nonce, expires_after)
+    envelope = {
+        "multiSigActionHash": multi_sig_action_hash,
+        "nonce": nonce,
+    }
+    return sign_user_signed_action(
+        wallet,
+        envelope,
+        MULTI_SIG_ENVELOPE_SIGN_TYPES,
+        "HyperliquidTransaction:SendMultiSig",
+        is_mainnet,
+    )
+
+
 def sign_usd_transfer_action(wallet, action, is_mainnet):
     return sign_user_signed_action(
         wallet,
         action,
-        [
-            {"name": "hyperliquidChain", "type": "string"},
-            {"name": "destination", "type": "string"},
-            {"name": "amount", "type": "string"},
-            {"name": "time", "type": "uint64"},
-        ],
+        USD_SEND_SIGN_TYPES,
         "HyperliquidTransaction:UsdSend",
         is_mainnet,
     )
@@ -180,13 +315,7 @@ def sign_spot_transfer_action(wallet, action, is_mainnet):
     return sign_user_signed_action(
         wallet,
         action,
-        [
-            {"name": "hyperliquidChain", "type": "string"},
-            {"name": "destination", "type": "string"},
-            {"name": "token", "type": "string"},
-            {"name": "amount", "type": "string"},
-            {"name": "time", "type": "uint64"},
-        ],
+        SPOT_TRANSFER_SIGN_TYPES,
         "HyperliquidTransaction:SpotSend",
         is_mainnet,
     )
@@ -196,12 +325,7 @@ def sign_withdraw_from_bridge_action(wallet, action, is_mainnet):
     return sign_user_signed_action(
         wallet,
         action,
-        [
-            {"name": "hyperliquidChain", "type": "string"},
-            {"name": "destination", "type": "string"},
-            {"name": "amount", "type": "string"},
-            {"name": "time", "type": "uint64"},
-        ],
+        WITHDRAW_SIGN_TYPES,
         "HyperliquidTransaction:Withdraw",
         is_mainnet,
     )
@@ -211,12 +335,7 @@ def sign_usd_class_transfer_action(wallet, action, is_mainnet):
     return sign_user_signed_action(
         wallet,
         action,
-        [
-            {"name": "hyperliquidChain", "type": "string"},
-            {"name": "amount", "type": "string"},
-            {"name": "toPerp", "type": "bool"},
-            {"name": "nonce", "type": "uint64"},
-        ],
+        USD_CLASS_TRANSFER_SIGN_TYPES,
         "HyperliquidTransaction:UsdClassTransfer",
         is_mainnet,
     )
@@ -226,31 +345,12 @@ def sign_convert_to_multi_sig_user_action(wallet, action, is_mainnet):
     return sign_user_signed_action(
         wallet,
         action,
-        [
-            {"name": "hyperliquidChain", "type": "string"},
-            {"name": "signers", "type": "string"},
-            {"name": "nonce", "type": "uint64"},
-        ],
+        CONVERT_TO_MULTI_SIG_USER_SIGN_TYPES,
         "HyperliquidTransaction:ConvertToMultiSigUser",
         is_mainnet,
     )
 
 
-def sign_convert_to_multi_sig_signer_action(wallet, action, is_mainnet):
-    return sign_user_signed_action(
-        wallet,
-        action,
-        [
-            {"name": "hyperliquidChain", "type": "string"},
-            {"name": "signer", "type": "address"},
-            {"name": "multiSigUser", "type": "address"},
-            {"name": "nonce", "type": "uint64"},
-        ],
-        "HyperliquidTransaction:ConvertToMultiSigSigner",
-        is_mainnet,
-    )
-
-
 def sign_agent(wallet, action, is_mainnet):
     return sign_user_signed_action(
         wallet,
@@ -281,14 +381,40 @@ def sign_approve_builder_fee(wallet, action, is_mainnet):
     )
 
 
+def sign_token_delegate_action(wallet, action, is_mainnet):
+    return sign_user_signed_action(
+        wallet,
+        action,
+        TOKEN_DELEGATE_TYPES,
+        "HyperliquidTransaction:TokenDelegate",
+        is_mainnet,
+    )
+
+
 def sign_inner(wallet, data):
-    structured_data = encode_structured_data(data)
-    signed = wallet.sign_message(structured_data)
+    signed = wallet.sign_typed_data(full_message=data)
     return {"r": to_hex(signed["r"]), "s": to_hex(signed["s"]), "v": signed["v"]}
 
 
+def recover_agent_or_user_from_l1_action(action, signature, active_pool, nonce, expires_after, is_mainnet):
+    hash = action_hash(action, active_pool, nonce, expires_after)
+    phantom_agent = construct_phantom_agent(hash, is_mainnet)
+    data = l1_payload(phantom_agent)
+    structured_data = encode_typed_data(full_message=data)
+    address = Account.recover_message(structured_data, vrs=[signature["v"], signature["r"], signature["s"]])
+    return address
+
+
+def recover_user_from_user_signed_action(action, signature, payload_types, primary_type, is_mainnet):
+    action["hyperliquidChain"] = "Mainnet" if is_mainnet else "Testnet"
+    data = user_signed_payload(primary_type, payload_types, action)
+    structured_data = encode_typed_data(full_message=data)
+    address = Account.recover_message(structured_data, vrs=[signature["v"], signature["r"], signature["s"]])
+    return address
+
+
 def float_to_wire(x: float) -> str:
-    rounded = "{:.8f}".format(x)
+    rounded = f"{x:.8f}"
     if abs(float(rounded) - x) >= 1e-12:
         raise ValueError("float_to_wire causes rounding", x)
     if rounded == "-0":
